# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import os
import json
import logging
import random
import string
import sys

import pytest
import requests
from influxdb import InfluxDBClient
from influxdb.exceptions import InfluxDBClientError
from pyinfluxql import Query
from grafana_api_client import GrafanaClient, GrafanaClientError
from twisted.internet import reactor
from twisted.internet.task import deferLater

import kotori


logger = logging.getLogger(__name__)


def boot_kotori(config):
    options = {
        '--config': config,
        '--debug': True,
        '--debug_io': True,
        '--debug_mqtt': True,
        '--debug_influx': True,
    }
    loader = kotori.boot(options)
    return loader


class GrafanaWrapper:

    def __init__(self, settings):
        self.settings = settings
        self.client = self.get_client()

    def get_client(self):
        return GrafanaClient((self.settings.grafana_username, self.settings.grafana_password), host='localhost', port=3000)

    def make_reset(self):

        @pytest.fixture(scope="function")
        def reset_grafana():
            """
            Fixture to delete the Grafana datasource and dashboard.
            """

            logger.info('Grafana: Resetting artefacts')

            for datasource in self.client.datasources.get():
                if datasource['name'] == self.settings.influx_database:
                    datasource_id = datasource['id']
                    self.client.datasources[datasource_id].delete()
                    break

            for dashboard_name in self.settings.grafana_dashboards:
                try:
                    self.client.dashboards.db[dashboard_name].delete()
                except GrafanaClientError as ex:
                    if '404' not in str(ex):
                        raise

        return reset_grafana


class InfluxWrapper:

    def __init__(self, database, measurement):
        self.database = database
        self.measurement = measurement
        self.client = self.get_client()
        self.create = None
        self.reset = None

    def get_client(self):
        # TODO: Get configuration parameters from .ini file or from runtime application.
        influx_client = InfluxDBClient(
            host='localhost',
            database=self.database,
            timeout=2)
        return influx_client

    def query(self):
        logger.info('InfluxDB: Querying database')
        expression = Query('*').from_(self.measurement) # .where(time__gte=time_begin, time__lte=time_end, **tags)
        influx_client = self.get_client()
        result = influx_client.query(str(expression))
        return result

    def get_record(self, index=None):
        result = self.query()

        # One measurement?
        assert len(result) == 1, "No data in database: len(result) = {}".format(len(result))

        # One record.
        records = list(result[self.measurement])
        assert len(records) >= 1

        # Check record.
        record = records[index]

        return record

    def get_first_record(self):
        return self.get_record(index=0)

    def make_create_db(self):
        @pytest.fixture(scope="package")
        def create():
            logger.info('InfluxDB: Creating database')
            #self.client.drop_database(self.database)
            self.client.create_database(self.database)
        return create

    def make_reset_measurement(self):

        @pytest.fixture(scope="function")
        def reset_measurement():
            logger.info('InfluxDB: Resetting database')
            # Clear out the database.
            influx = InfluxWrapper(database=self.database, measurement=self.measurement)
            influx.client.delete_series(self.database, measurement=self.measurement)
            #try:
            #except InfluxDBClientError as ex:
            #    if 'database not found: mqttkit_1_itest' not in str(ex):
            #        raise

        return reset_measurement


def sleep(secs):
    # https://gist.github.com/jhorman/891717
    return deferLater(reactor, secs, lambda: None)


def mqtt_json_sensor(topic, data):
    payload = json.dumps(data)
    return mqtt_sensor(topic, payload)


def mqtt_sensor(topic, payload):

    logger.info('MQTT: Submitting reading')

    # When running on CI (GHA), run ``mosquitto_pub`` from Docker image.
    # https://stackoverflow.com/questions/24319662
    if os.environ.get("CI"):
        if sys.platform == "linux":
            mosquitto_pub = "docker run --rm --network=host eclipse-mosquitto:1.6.12 mosquitto_pub -h localhost"
        elif sys.platform == "darwin":
            mosquitto_pub = "docker run --rm eclipse-mosquitto:1.6.12 mosquitto_pub -h host.docker.internal"
        else:
            raise NotImplementedError("Invoking 'mosquitto_pub' through Docker on '{}' not supported yet".format(sys.platform))
    else:
        mosquitto_pub = "mosquitto_pub -h localhost"
    command = "{mosquitto_pub} -t '{topic}' -m '{payload}'".format(mosquitto_pub=mosquitto_pub, topic=topic, payload=payload)

    logger.info('Running command {}'.format(command))
    exitcode = os.system(command)
    if exitcode != 0:
        raise ChildProcessError("Invoking command failed: {command}. Exit code: {exitcode}".format(command=command, exitcode=exitcode))


def http_json_sensor(topic, data):
    uri = 'http://localhost:24642/api{}'.format(topic)
    logger.info('HTTP: Submitting reading to {} using JSON'.format(uri))
    requests.post(uri, json=data)


def http_form_sensor(topic, data):
    uri = 'http://localhost:24642/api{}'.format(topic)
    logger.info('HTTP: Submitting reading to {} using x-www-form-urlencoded'.format(uri))
    requests.post(uri, data=data)


def http_csv_sensor(topic, data):
    uri = 'http://localhost:24642/api{}'.format(topic)
    logger.info('HTTP: Submitting reading to {} using CSV'.format(uri))
    body = ''
    body += '## {}\n'.format(','.join(map(str, list(data.keys()))))
    body += '{}\n'.format(','.join(map(str, list(data.values()))))
    requests.post(uri, data=body, headers={'Content-Type': 'text/csv'})


def http_get_data(topic=None, format='csv', ts_from=None, ts_to=None):
    uri = 'http://localhost:24642/api{topic}.{format}?from={ts_from}&to={ts_to}'.format(
        topic=topic, format=format, ts_from=ts_from, ts_to=ts_to)
    logger.info('HTTP: Exporting data from {} using format "{}"'.format(uri, format))
    payload = requests.get(uri).content
    if format in ["csv", "txt", "json", "html"]:
        payload = payload.decode()
    return payload


def idgen(size=6, chars=string.ascii_uppercase + string.digits):
    """
    ========
    Synopsis
    ========
    ```
    >>> idgen()
    'G5G74W'
    ```

    - https://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits/2257449#2257449
    """
    return ''.join(random.choice(chars) for _ in range(size))



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging

import pytest
import pytest_twisted
from bunch import Bunch

from test.conftest import create_machinery
from test.settings.mqttkit import PROCESS_DELAY_MQTT
from test.util import mqtt_json_sensor, sleep, InfluxWrapper, GrafanaWrapper

logger = logging.getLogger(__name__)


settings = Bunch(
    influx_database='hiveeyes_itest',
    influx_measurement='site_box_sensors',
    mqtt_topic='hiveeyes/itest/site/box/data.json',
    grafana_username='admin',
    grafana_password='admin',
    grafana_dashboards=['hiveeyes-itest-site-box', 'hiveeyes-itest'],
)

influx = InfluxWrapper(database=settings.influx_database, measurement=settings.influx_measurement)
grafana = GrafanaWrapper(settings=settings)

machinery_hiveeyes = create_machinery('./etc/test/hiveeyes.ini')
create_influxdb_hiveeyes = influx.make_create_db()
reset_influxdb_hiveeyes = influx.make_reset_measurement()
reset_grafana_hiveeyes = grafana.make_reset()


@pytest_twisted.inlineCallbacks
@pytest.mark.hiveeyes
def test_mqtt_to_grafana(machinery_hiveeyes, create_influxdb_hiveeyes, reset_influxdb_hiveeyes, reset_grafana_hiveeyes):
    """
    Publish a single reading in JSON format to MQTT and proof
    - it is stored in the InfluxDB database.
    - a corresponding datasource and dashboards have been created in Grafana.
    """

    # Submit a single measurement, without timestamp.
    data = {
        'temperature': 42.84,
        'weight': 33.33,
    }
    yield mqtt_json_sensor(settings.mqtt_topic, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Wait for Grafana to create its artefacts.
    yield sleep(2)

    # Proof that data arrived in InfluxDB.
    record = influx.get_first_record()
    del record['time']
    assert record == {u'temperature': 42.84, u'weight': 33.33}
    yield record

    # Proof that Grafana is well provisioned.
    logger.info('Grafana: Checking datasource')
    datasource_names = []
    for datasource in grafana.client.datasources.get():
        datasource_names.append(datasource['name'])
    assert settings.influx_database in datasource_names

    logger.info('Grafana: Checking dashboards')
    for dashboard_name in settings.grafana_dashboards:
        dashboard = grafana.client.dashboards.db[dashboard_name].get()['dashboard']
        if 'rows' in dashboard:
            umbrella = dashboard['rows'][0]
        else:
            umbrella = dashboard
        target = umbrella['panels'][0]['targets'][0]
        #assert target['measurement'] == settings.influx_measurement
        assert 'temperature' in target['query'] or 'weight' in target['query']



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging

import pytest
import pytest_twisted
from twisted.internet import threads

from test.settings.mqttkit import settings, influx_sensors, PROCESS_DELAY_HTTP
from test.util import http_json_sensor, http_form_sensor, http_csv_sensor, sleep

logger = logging.getLogger(__name__)


@pytest_twisted.inlineCallbacks
@pytest.mark.http
def test_http_json(machinery, create_influxdb, reset_influxdb):
    """
    Submit single reading in JSON format to HTTP API
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single measurement, without timestamp.
    data = {
        'temperature': 25.26,
        'humidity': 51.8,
    }
    yield threads.deferToThread(http_json_sensor, settings.channel_path_data, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_HTTP)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    del record['time']
    assert record == {u'temperature': 25.26, u'humidity': 51.8}
    yield record


@pytest_twisted.inlineCallbacks
@pytest.mark.http
def test_http_urlencoded(machinery, create_influxdb, reset_influxdb):
    """
    Submit single reading in ``x-www-form-urlencoded`` format to HTTP API
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single measurement, without timestamp.
    data = {
        'temperature': 25.26,
        'humidity': 51.8,
    }
    yield threads.deferToThread(http_form_sensor, settings.channel_path_data, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_HTTP)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    del record['time']
    assert record == {u'temperature': 25.26, u'humidity': 51.8}
    yield record


@pytest_twisted.inlineCallbacks
@pytest.mark.http
@pytest.mark.mongodb
def test_http_csv(machinery, create_influxdb, reset_influxdb):
    """
    Submit single reading in CSV format to HTTP API
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single measurement, without timestamp.
    data = {
        'temperature': 25.26,
        'humidity': 51.8,
    }
    yield threads.deferToThread(http_csv_sensor, settings.channel_path_data, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_HTTP)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    del record['time']
    assert record == {u'temperature': 25.26, u'humidity': 51.8}
    yield record



# -*- coding: utf-8 -*-
# (c) 2015-2021 Andreas Motl, <andreas@getkotori.org>
from kotori.daq.decoder import MessageType


class StrategyBase:

    @staticmethod
    def sanitize_db_identifier(value):
        """
        Different databases accept different special
        characters as database- or table names.

        Better safe than sorry, let's strip them all.
        """
        value = value.replace('/', '_').replace('.', '_').replace('-', '_').lower()
        return value

    @staticmethod
    def get_table_suffix(topology, message_type=None):
        """
        Derive database table suffix from message type.
        """
        if message_type in (MessageType.DATA_CONTAINER, MessageType.DATA_DISCRETE) or topology.slot.startswith('data'):
            suffix = 'sensors'
        elif message_type == MessageType.EVENT or topology.slot.startswith('event'):
            suffix = 'events'
        else:
            suffix = 'unknown'
        return suffix



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging

import pytest
import pytest_twisted
from twisted.internet import threads

from test.settings.basic import settings, influx_sensors, grafana, create_influxdb, reset_influxdb, reset_grafana, PROCESS_DELAY_MQTT
from test.util import http_json_sensor, sleep

logger = logging.getLogger(__name__)


# https://community.hiveeyes.org/t/more-data-acquisition-payload-formats-for-kotori/1421/2


data_in = {
  "esp8266id": 12041741,
  "sensordatavalues": [
    {
      "value_type": "SDS_P1",
      "value": "35.67"
    },
    {
      "value_type": "SDS_P2",
      "value": "17.00"
    },
    {
      "value_type": "BME280_temperature",
      "value": "-2.83"
    },
    {
      "value_type": "BME280_humidity",
      "value": "66.73"
    },
    {
      "value_type": "BME280_pressure",
      "value": "100535.97"
    },
    {
      "value_type": "samples",
      "value": "3016882"
    },
    {
      "value_type": "min_micro",
      "value": "77"
    },
    {
      "value_type": "max_micro",
      "value": "26303"
    },
    {
      "value_type": "signal",
      "value": "-66"
    }
  ],
  "software_version": "NRZ-2018-123B"
}


data_out = {
  u'SDS_P1': 35.67,
  u'SDS_P2': 17.00,
  u'BME280_temperature': -2.83,
  u'BME280_humidity': 66.73,
  u'BME280_pressure': 100535.97,
  u'samples': 3016882,
  u'min_micro': 77,
  u'max_micro': 26303,
  u'signal': -66,
}


create_influxdb = influx_sensors.make_create_db()
reset_influxdb = influx_sensors.make_reset_measurement()


@pytest_twisted.inlineCallbacks
@pytest.mark.http
@pytest.mark.airrohr
def test_airrohr_http_json(machinery, create_influxdb, reset_influxdb):
    """
    Submit single reading in Airrohr JSON format to HTTP API
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single measurement, without timestamp.
    yield threads.deferToThread(http_json_sensor, settings.channel_path_airrohr, data_in)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    del record['time']
    assert record == data_out
    yield record



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging

import pytest
import pytest_twisted

from test.settings.mqttkit import settings, grafana, PROCESS_DELAY_MQTT
from test.util import mqtt_json_sensor, sleep

logger = logging.getLogger(__name__)


@pytest_twisted.inlineCallbacks
@pytest.mark.grafana
def test_mqtt_to_grafana(machinery, create_influxdb, reset_influxdb, reset_grafana):
    """
    Publish single reading in JSON format to MQTT broker and proof
    that a corresponding datasource and a dashboard was created in Grafana.
    """

    # Submit a single measurement, without timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
    }
    yield mqtt_json_sensor(settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)
    yield sleep(PROCESS_DELAY_MQTT)
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that Grafana is well provisioned.
    logger.info('Grafana: Checking datasource')
    datasource_names = []
    for datasource in grafana.client.datasources.get():
        datasource_names.append(datasource['name'])
    assert settings.influx_database in datasource_names

    logger.info('Grafana: Checking dashboard')
    dashboard_name = settings.grafana_dashboards[0]
    dashboard = grafana.client.dashboards.db[dashboard_name].get()
    target = dashboard['dashboard']['rows'][0]['panels'][0]['targets'][0]
    assert target['measurement'] == settings.influx_measurement_sensors
    assert 'temperature' in target['query'] or 'humidity' in target['query']



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import json
from collections import OrderedDict


class AirrohrDecoder:
    """
    Decode JSON payloads in Airrohr format.

    Documentation
    =============
    - https://getkotori.org/docs/handbook/decoders/airrohr.html
    - https://community.hiveeyes.org/t/more-data-acquisition-payload-formats-for-kotori/1421/2

    Example
    =======
    ::

        {
          "esp8266id": 12041741,
          "sensordatavalues": [
            {
              "value_type": "SDS_P1",
              "value": "35.67"
            },
            {
              "value_type": "SDS_P2",
              "value": "17.00"
            },
            {
              "value_type": "BME280_temperature",
              "value": "-2.83"
            },
            {
              "value_type": "BME280_humidity",
              "value": "66.73"
            },
            {
              "value_type": "BME280_pressure",
              "value": "100535.97"
            },
            {
              "value_type": "samples",
              "value": "3016882"
            },
            {
              "value_type": "min_micro",
              "value": "77"
            },
            {
              "value_type": "max_micro",
              "value": "26303"
            },
            {
              "value_type": "signal",
              "value": "-66"
            }
          ],
          "software_version": "NRZ-2018-123B"
        }

    """

    @staticmethod
    def decode(payload):

        # Decode from JSON.
        message = json.loads(payload)

        # Create data dictionary by flattening nested message.
        data = OrderedDict()
        for item in message.get('sensordatavalues', []):
            key = item['value_type']
            value = item['value']
            data[key] = value

        return data



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging
from copy import deepcopy

import pytest
import pytest_twisted
from twisted.internet import threads

from test.settings.mqttkit import settings, influx_sensors, PROCESS_DELAY_MQTT
from test.util import mqtt_json_sensor, sleep, idgen

logger = logging.getLogger(__name__)


@pytest_twisted.inlineCallbacks
@pytest.mark.influxdb
def test_influxdb_tags(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker.
    Proof that all special fields are stored in the
    InfluxDB database as tags.
    """

    # Define field names which are tags.
    # FIXME: Synchronize with ``kotori.daq.influx.storage.format_chunk()``.
    tag_fields_main = ['geohash', 'latitude', 'longitude']
    tag_fields_more = ['location', 'location_id', 'location_name', 'sensor_id', 'sensor_type']

    # Submit a single measurement, without timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
        'geohash': 'aa3434',
        'latitude': 42.50,
        'longitude': 52.40,
    }

    for field in tag_fields_more:
        data[field] = idgen()

    yield threads.deferToThread(mqtt_json_sensor, settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Define reference payload.
    reference = deepcopy(data)
    reference['latitude'] = str(reference['latitude'])
    reference['longitude'] = str(reference['longitude'])

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    del record['time']
    assert record == reference
    yield record

    # Proof that data all special fields have been converged to tags.
    resultset = influx_sensors.client.query('SHOW TAG KEYS FROM {measurement};'.format(measurement=settings.influx_measurement_sensors))
    yield resultset

    tag_names = []
    for item in resultset[settings.influx_measurement_sensors]:
        tag_names.append(item['tagKey'])

    assert set(tag_names) == set(tag_fields_main + tag_fields_more)



# -*- coding: utf-8 -*-
# (c) 2015-2021 Andreas Motl <andreas@getkotori.org>
import requests
from copy import deepcopy
from funcy import project
from collections import OrderedDict
from twisted.logger import Logger
from influxdb.client import InfluxDBClient, InfluxDBClientError
from kotori.io.protocol.util import parse_timestamp, is_number, convert_floats

log = Logger()


class InfluxDBAdapter(object):

    def __init__(self, settings=None, database=None):

        settings = deepcopy(settings) or {}
        settings.setdefault('host', u'localhost')
        settings.setdefault('port', u'8086')
        settings.setdefault('username', u'root')
        settings.setdefault('password', u'root')
        settings.setdefault('database', database)
        settings.setdefault('pool_size', 10)

        settings.setdefault('use_udp', False)
        settings.setdefault('udp_port', u'4444')

        settings['port'] = int(settings['port'])
        settings['udp_port'] = int(settings['udp_port'])

        self.__dict__.update(**settings)

        # Bookeeping for all databases having been touched already
        self.databases_written_once = set()

        # Knowledge about all databases to be accessed using UDP
        # TODO: Refactor to configuration setting
        self.udp_databases = [
            {'name': 'luftdaten_info', 'port': u'4445'},
        ]
        self.host_uri = u'influxdb://{host}:{port}'.format(**self.__dict__)

        log.info(u'Storage target is {uri}, pool size is {pool_size}', uri=self.host_uri, pool_size=self.pool_size)
        self.influx_client = InfluxDBClient(
            host=self.host, port=self.port,
            username=self.username, password=self.password,
            database=self.database, pool_size=self.pool_size,
            timeout=10)

        # TODO: Hold references to multiple UDP databases using mapping "self.udp_databases".
        self.influx_client_udp = None
        if settings['use_udp']:
            self.influx_client_udp = InfluxDBClient(
                host=self.host, port=self.port,
                username=self.username, password=self.password,
                use_udp=settings['use_udp'], udp_port=settings['udp_port'],
                timeout=10)

    def is_udp_database(self, name):
        for entry in self.udp_databases:
            if entry['name'] == name:
                return True
        return False

    def write(self, meta, data):

        meta_copy = deepcopy(dict(meta))
        data_copy = deepcopy(data)

        try:
            chunk = self.format_chunk(meta, data)

        except Exception as ex:
            log.failure(u'Could not format chunk (ex={ex_name}: {ex}): data={data}, meta={meta}',
                ex_name=ex.__class__.__name__, ex=ex, meta=meta_copy, data=data_copy)
            raise

        try:
            success = self.write_chunk(meta, chunk)
            return success

        except requests.exceptions.ConnectionError as ex:
            log.failure(u'Problem connecting to InfluxDB at {uri}: {ex}', uri=self.host_uri, ex=ex)
            raise

        except InfluxDBClientError as ex:

            if ex.code == 404 or 'database not found' in ex.content:

                log.info('Creating database "{database}"', database=meta.database)
                self.influx_client.create_database(meta.database)

                # Attempt second write
                success = self.write_chunk(meta, chunk)
                return success

                #log.failure('InfluxDBClientError: {ex}', ex=ex)

            # [0.8] ignore "409: database kotori-dev exists"
            # [0.9] ignore "database already exists"
            elif ex.code == 409 or 'database already exists' in ex.content:
                pass
            else:
                raise

    def write_chunk(self, meta, chunk):
        if self.influx_client_udp and self.is_udp_database(meta.database) and meta.database in self.databases_written_once:
            success = self.influx_client_udp.write_points([chunk], time_precision='s', database=meta.database)
        else:
            success = self.influx_client.write_points([chunk], time_precision=chunk['time_precision'], database=meta.database)
            self.databases_written_once.add(meta.database)
        if success:
            log.debug(u"Storage success: {chunk}", chunk=chunk)
        else:
            log.error(u"Storage failed:  {chunk}", chunk=chunk)
        return success

    @staticmethod
    def get_tags(data):
        return project(data, ['gateway', 'node'])

    def format_chunk(self, meta, data):
        """
        Format for InfluxDB >= 0.9::
        {
            "measurement": "hiveeyes_100",
            "tags": {
                "host": "server01",
                "region": "europe"
            },
            "time": "2015-10-17T19:30:00Z",
            "fields": {
                "value": 0.42
            }
        }
        """

        assert isinstance(data, dict), 'Data payload is not a dictionary'

        chunk = {
            "measurement": meta['measurement'],
            "tags": {},
        }

        """
        if "gateway" in meta:
            chunk["tags"]["gateway"] = meta["gateway"]

        if "node" in meta:
            chunk["tags"]["node"]    = meta["node"]
        """

        # TODO: Refactor to some knowledgebase component.
        time_field_candidates = [
            'time',         # Vanilla
            'datetime',     # Vanilla
            'Time',         # Tasmota
            'dateTime',     # WeeWX
            'timestamp',    # Contrib
        ]

        # Extract timestamp field from data
        chunk['time_precision'] = 'n'
        # FIXME: Unify with ``kotori.io.protocol.http.data_acquisition()``.
        for time_field in time_field_candidates:
            if time_field in data:

                # WeeWX. TODO: Move to specific vendor configuration.
                # Disabled in favor of precision detection heuristic.
                #if time_field == 'dateTime':
                #    chunk['time_precision'] = 's'

                # Process timestamp field.
                if data[time_field]:

                    # Decode timestamp.
                    chunk['time'] = data[time_field]
                    if is_number(chunk['time']):
                        chunk['time'] = int(float(chunk['time']))

                    # Remove timestamp from data payload.
                    del data[time_field]

                    # If we found a timestamp field already,
                    # don't look out for more.
                    break

        # Extract geohash from data. Finally, thanks Rich!
        # TODO: Also precompute geohash with 3-4 different zoomlevels and add them as tags
        if "geohash" in data:
            chunk["tags"]["geohash"] = data["geohash"]
            del data['geohash']

        if "latitude" in data and "longitude" in data:
            chunk["tags"]["latitude"] = data["latitude"]
            chunk["tags"]["longitude"] = data["longitude"]
            del data['latitude']
            del data['longitude']

        # Extract more information specific to luftdaten.info
        for field in ['location', 'location_id', 'location_name', 'sensor_id', 'sensor_type']:
            if field in data:
                chunk["tags"][field] = data[field]
                del data[field]

        # TODO: Maybe do this at data acquisition / transformation time, not here.
        if 'time' in chunk:
            timestamp = chunk['time'] = parse_timestamp(chunk['time'])

            # Heuristically compute timestamp precision
            if isinstance(timestamp, int):
                if timestamp >= 1e17 or timestamp <= -1e17:
                    time_precision = 'n'
                elif timestamp >= 1e14 or timestamp <= -1e14:
                    time_precision = 'u'
                elif timestamp >= 1e11 or timestamp <= -1e11:
                    time_precision = 'ms'

                # FIXME: Is this a reasonable default?
                else:
                    time_precision = 's'

                chunk['time_precision'] = time_precision

            """
            # FIXME: Breaks CSV data acquisition. Why?
            if isinstance(chunk['time'], datetime.datetime):
                if chunk['time'].microsecond == 0:
                    chunk['time_precision'] = 's'
            """

        """
        Prevent errors like
        ERROR: InfluxDBClientError: 400:
                       write failed: field type conflict:
                       input field "pitch" on measurement "01_position" is type float64, already exists as type integer
        """
        self.data_to_float(data)

        assert data, 'Data payload is empty'

        chunk["fields"] = data

        return chunk

    def data_to_float(self, data):
        return convert_floats(data)

        for key, value in data.items():

            # Sanity checks
            if isinstance(value, str):
                continue

            if value is None:
                data[key] = None
                continue

            # Convert to float
            try:
                data[key] = float(value)
            except (TypeError, ValueError) as ex:
                log.warn(u'Measurement "{key}: {value}" float conversion failed: {ex}', key=key, value=value, ex=ex)


class BusInfluxForwarder(object):
    """
    Generic software bus -> influxdb forwarder based on prototypic implementation at HiveEyes
    TODO: Generalize and refactor
    """

    # TODO: Improve parameter passing
    def __init__(self, bus, topic, config, channel):
        self.bus = bus
        self.topic = topic
        self.config = config
        self.channel = channel

        self.bus.subscribe(self.bus_receive, self.topic)


    def topic_to_topology(self, topic):
        raise NotImplementedError()

    @staticmethod
    def sanitize_db_identifier(value):
        value = str(value).replace('/', '_').replace('.', '_').replace('-', '_')
        return value

    def bus_receive(self, payload):
        try:
            return self.process_message(self.topic, payload)
        except Exception:
            log.failure(u'Processing bus message failed')

    def process_message(self, topic, payload, *args):

        log.debug('Bus receive: topic={topic}, payload={payload}', topic=topic, payload=payload)

        # TODO: filter by realm/topic

        # decode message
        if isinstance(payload, dict):
            message = payload.copy()
        elif isinstance(payload, list):
            message = OrderedDict(payload)
        else:
            raise TypeError('Unable to handle data type "{}" from bus'.format(type(payload)))

        # compute storage location from topic and message
        storage_location = self.storage_location(message)
        log.debug('Storage location: {storage_location}', storage_location=dict(storage_location))

        # store data
        self.store_message(storage_location, message)

    def storage_location(self, data):
        raise NotImplementedError()

    def store_encode(self, data):
        return data

    def store_message(self, location, data):

        data = self.store_encode(data)

        influx = InfluxDBAdapter(
            settings = self.config['influxdb'],
            database = location.database)

        outcome = influx.write(location, data)
        log.debug('Store outcome: {outcome}', outcome=outcome)

        self.on_store(location, data)

    def on_store(self, location, data):
        pass




# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging

import pytest
import pytest_twisted
from twisted.internet import threads

from test.settings.mqttkit import settings, influx_sensors, PROCESS_DELAY_MQTT
from test.util import mqtt_json_sensor, sleep

logger = logging.getLogger(__name__)


# The WeeWX decoder listens on the special MQTT topic suffix "/loop".
mqtt_topic = 'mqttkit-1/itest/foo/bar/loop'

# Data payload captured from a Vantage Pro 2 weather station.
data = {
    "windSpeed10_kph": "5.78725803977",
    "monthET": "1.32",
    "highUV": "0.0",
    "cloudbase_meter": "773.082217509",
    "leafTemp1_C": "8.33333333333",
    "rainAlarm": "0.0",
    "pressure_mbar": "948.046280104",
    "rain_cm": "0.0",
    "highRadiation": "0.0",
    "interval_minute": "5.0",
    "barometer_mbar": "1018.35464712",
    "yearRain_cm": "17.2000000043",
    "consBatteryVoltage_volt": "4.72",
    "dewpoint_C": "2.07088485785",
    "insideAlarm": "0.0",
    "inHumidity": "29.0",
    "soilLeafAlarm4": "0.0",
    "sunrise": "1492489200.0",
    "windGust_kph": "9.65608800006",
    "heatindex_C": "3.55555555556",
    "dayRain_cm": "0.0",
    "lowOutTemp": "38.3",
    "outsideAlarm1": "0.0",
    "forecastIcon": "8.0",
    "outsideAlarm2": "0.0",
    "windSpeed_kph": "3.95409343049",
    "forecastRule": "40.0",
    "windrun_km": "1.07449640224",
    "outHumidity": "90.0",
    "stormStart": "1492207200.0",
    "inDewpoint": "45.1231125123",
    "altimeter_mbar": "1016.62778614",
    "windchill_C": "3.55555555556",
    "appTemp_C": "1.26842313302",
    "outTemp_C": "3.55555555556",
    "windGustDir": "275.0",
    "extraAlarm1": "0.0",
    "extraAlarm2": "0.0",
    "extraAlarm3": "0.0",
    "extraAlarm4": "0.0",
    "extraAlarm5": "0.0",
    "extraAlarm6": "0.0",
    "extraAlarm7": "0.0",
    "extraAlarm8": "0.0",
    "humidex_C": "3.55555555556",
    "rain24_cm": "0.88000000022",
    "rxCheckPercent": "87.9791666667",
    "hourRain_cm": "0.0",
    "inTemp_C": "26.8333333333",
    "watertemp": "8.33333333333",
    "trendIcon": "59.7350993377",
    "soilLeafAlarm2": "0.0",
    "soilLeafAlarm3": "0.0",
    "usUnits": "16.0",
    "soilLeafAlarm1": "0.0",
    "leafWet4": "0.0",
    "txBatteryStatus": "0.0",
    "yearET": "4.88",
    "monthRain_cm": "2.94000000074",
    "UV": "0.0",
    "rainRate_cm_per_hour": "0.0",
    "dayET": "0.0",
    "dateTime": "1492467300.0",
    "windDir": "283.55437192",
    "stormRain_cm": "1.72000000043",
    "ET_cm": "0.0",
    "sunset": "1492538940.0",
    "highOutTemp": "38.4",
    "radiation_Wpm2": "0.0"
}


@pytest_twisted.inlineCallbacks
@pytest.mark.mqtt
@pytest.mark.weewx
def test_weewx_mqtt(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single measurement, without timestamp.
    yield threads.deferToThread(mqtt_json_sensor, settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()

    assert record["time"] == '2017-04-17T22:15:00Z'

    assert record["outTemp_C"] == 3.55555555556
    assert record["windSpeed10_kph"] == 5.78725803977
    assert record["cloudbase_meter"] == 773.082217509
    assert record["consBatteryVoltage_volt"] == 4.72

    yield record



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging

import pytest
import pytest_twisted
from twisted.internet import threads

from test.settings.mqttkit import settings, influx_events, PROCESS_DELAY_MQTT
from test.util import mqtt_json_sensor, sleep, http_json_sensor, http_form_sensor

logger = logging.getLogger(__name__)

event_data = {
    'title': 'Some event',
    'text': '<a href="https://somewhere.example.org/events?reference=482a38ce-791e-11e6-b152-7cd1c55000be">see also</a>',
    'tags': 'event,alert,important',
    'reference': '482a38ce-791e-11e6-b152-7cd1c55000be',
}


@pytest_twisted.inlineCallbacks
@pytest.mark.mqtt
@pytest.mark.events
def test_event_mqtt(machinery, create_influxdb, reset_influxdb_events):
    """
    Publish event in JSON format to MQTT broker
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single event, without timestamp.
    yield threads.deferToThread(mqtt_json_sensor, settings.mqtt_topic_event, event_data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that event arrived in InfluxDB.
    record = influx_events.get_first_record()
    del record['time']
    assert record == event_data


@pytest_twisted.inlineCallbacks
@pytest.mark.http
@pytest.mark.events
def test_event_http_json(machinery, create_influxdb, reset_influxdb):
    """
    Submit event in JSON format to HTTP API
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single event, without timestamp.
    yield threads.deferToThread(http_json_sensor, settings.channel_path_event, event_data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that event arrived in InfluxDB.
    record = influx_events.get_first_record()
    del record['time']
    assert record == event_data
    yield record


@pytest_twisted.inlineCallbacks
@pytest.mark.http
@pytest.mark.events
def test_event_http_urlencoded(machinery, create_influxdb, reset_influxdb):
    """
    Submit event in ``x-www-form-urlencoded`` format to HTTP API
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single event, without timestamp.
    yield threads.deferToThread(http_form_sensor, settings.channel_path_event, event_data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that event arrived in InfluxDB.
    record = influx_events.get_first_record()
    del record['time']
    assert record == event_data
    yield record



# -*- coding: utf-8 -*-
# (c) 2019-2021 Andreas Motl <andreas@getkotori.org>
"""
Decode JSON payloads in Tasmota format.

Documentation
=============
- https://getkotori.org/docs/handbook/decoders/tasmota.html

Resources
=========
- https://github.com/arendst/tasmota
- https://github.com/arendst/tasmota/wiki/MQTT
- https://tasmota.github.io/docs/#/MQTT
"""
import json
from collections import OrderedDict
from copy import deepcopy


class TasmotaSensorDecoder:

    @staticmethod
    def decode(payload):
        """
        SonoffSC::

            {
              "Time": "2019-06-02T22:13:07",
              "SonoffSC": {
                "Temperature": 25,
                "Humidity": 15,
                "Light": 20,
                "Noise": 10,
                "AirQuality": 90
              },
              "TempUnit": "C"
            }

        Sonoff DS18B20::

            {
              "Time": "2017-02-16T10:13:52",
              "DS18B20": {
                "Temperature": 20.6
              }
            }

        Wemos Multi::

            {
              "Time": "2017-10-05T22:39:45",
              "DS18x20": {
                "DS1": {
                  "Type": "DS18B20",
                  "Address": "28FF4CBFA41604C4",
                  "Temperature": 25.37
                },
                "DS2": {
                  "Type": "DS18B20",
                  "Address": "28FF1E7FA116035D",
                  "Temperature": 30.44
                },
                "DS3": {
                  "Type": "DS18B20",
                  "Address": "28FF1597A41604CE",
                  "Temperature": 25.81
                }
              },
              "DHT22": {
                "Temperature": 33.2,
                "Humidity": 30
              },
              "TempUnit": "C"
            }

        """

        # Decode from JSON.
        message = json.loads(payload)

        # Decode message format.
        data = OrderedDict()

        # Transfer timestamp field.
        if 'Time' in message:
            data['Time'] = message['Time']

        # Transfer measurement fields.
        path = []
        for key, value in message.items():
            path.append(key)
            if isinstance(value, dict):
                for dkey, dvalue in value.items():
                    path.append(dkey)
                    if isinstance(dvalue, dict):
                        if 'Type' in dvalue:
                            subdata = deepcopy(dvalue)
                            del subdata['Type']
                            del subdata['Address']
                            for dskey, dsvalue in subdata.items():
                                path.append(dskey)
                                effective_key = '.'.join(path)
                                data[effective_key] = dsvalue
                                path.pop()
                    else:
                        effective_key = '.'.join(path)
                        data[effective_key] = dvalue
                    path.pop()

            path.pop()
        return data


class TasmotaStateDecoder:

    @staticmethod
    def decode(payload):
        """
        {
          "Time": "2019-06-02T22:13:07",
          "Uptime": "1T18:10:35",
          "Vcc": 3.182,
          "SleepMode": "Dynamic",
          "Sleep": 50,
          "LoadAvg": 19,
          "Wifi": {
            "AP": 1,
            "SSId": "{redacted}",
            "BSSId": "A0:F3:C1:{redacted}",
            "Channel": 1,
            "RSSI": 100,
            "LinkCount": 1,
            "Downtime": "0T00:00:07"
          }
        }
        """

        # Decode from JSON.
        message = json.loads(payload)

        # Decode message format.
        data = OrderedDict()
        data['Time'] = message.get('Time')
        data['Device.Vcc'] = message.get('Vcc')
        data['Device.Sleep'] = message.get('Sleep')
        data['Device.LoadAvg'] = message.get('LoadAvg')
        data['Device.Wifi.Channel'] = message.get('Wifi', {}).get('Channel')
        data['Device.Wifi.RSSI'] = message.get('Wifi', {}).get('RSSI')
        data['Device.Wifi.LinkCount'] = message.get('Wifi', {}).get('LinkCount')
        return data



# -*- coding: utf-8 -*-
# (c) 2020 Andreas Motl <andreas@getkotori.org>

from test.util import InfluxWrapper, GrafanaWrapper

PROCESS_DELAY_MQTT = 0.2
PROCESS_DELAY_HTTP = 0.2


class TestSettings:

    # InfluxDB settings.
    influx_database = 'mqttkit_1_itest'
    influx_measurement_sensors = 'foo_bar_sensors'
    influx_measurement_events = 'foo_bar_events'

    # Grafana settings.
    grafana_username = 'admin'
    grafana_password = 'admin'
    grafana_dashboards = ['mqttkit-1-itest']

    # MQTT channel settings.
    mqtt_topic_single = 'mqttkit-1/itest/foo/bar/data'
    mqtt_topic_json   = 'mqttkit-1/itest/foo/bar/data.json'
    mqtt_topic_event  = 'mqttkit-1/itest/foo/bar/event.json'
    mqtt_topic_homie  = 'mqttkit-1/itest/foo/bar/data/__json__'
    mqtt_topic_json_legacy = 'mqttkit-1/itest/foo/bar/message-json'

    # HTTP channel settings.
    channel_path_data    = '/mqttkit-1/itest/foo/bar/data'
    channel_path_event   = '/mqttkit-1/itest/foo/bar/event'
    channel_path_airrohr = '/mqttkit-1/itest/foo/bar/custom/airrohr'


settings = TestSettings

influx_sensors = InfluxWrapper(database=settings.influx_database, measurement=settings.influx_measurement_sensors)
influx_events = InfluxWrapper(database=settings.influx_database, measurement=settings.influx_measurement_events)
grafana = GrafanaWrapper(settings=settings)



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import json
import logging

import pytest
import pytest_twisted

from test.settings.mqttkit import influx_sensors, PROCESS_DELAY_MQTT
from test.util import mqtt_json_sensor, sleep

logger = logging.getLogger(__name__)


tasmota_sensor_topic = 'mqttkit-1/itest/foo/bar/tele/SENSOR'
tasmota_state_topic = 'mqttkit-1/itest/foo/bar/tele/STATE'


@pytest_twisted.inlineCallbacks
@pytest.mark.tasmota
def test_tasmota_sonoff_sc(machinery, create_influxdb, reset_influxdb):
    """
    Publish a single SENSOR reading in Tasmota/JSON format
    to MQTT broker, including a timestamp.
    Proof that the reading is processed and stored correctly.

    https://getkotori.org/docs/handbook/decoders/tasmota.html#submit-example-payload
    """

    # Submit a single measurement.
    data = {
      "Time": "2019-06-02T22:13:07",
      "SonoffSC": {
        "Temperature": 25,
        "Humidity": 15,
        "Light": 20,
        "Noise": 10,
        "AirQuality": 90
      },
      "TempUnit": "C"
    }
    yield mqtt_json_sensor(tasmota_sensor_topic, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Define reference data.
    reference = {
      u'time': u'2019-06-02T20:13:07Z',
      u'SonoffSC.AirQuality': 90,
      u'SonoffSC.Humidity': 15,
      u'SonoffSC.Light': 20,
      u'SonoffSC.Noise': 10,
      u'SonoffSC.Temperature': 25,
    }

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == reference
    yield record


@pytest_twisted.inlineCallbacks
@pytest.mark.tasmota
def test_tasmota_ds18b20(machinery, create_influxdb, reset_influxdb):
    """
    Publish another single SENSOR reading in Tasmota/JSON format
    to MQTT broker, including a timestamp.
    Proof that the reading is processed and stored correctly.

    https://getkotori.org/docs/handbook/decoders/tasmota.html#submit-example-payload
    """

    # Submit a single measurement.
    data = {
      "Time": "2017-02-16T10:13:52",
      "DS18B20": {
        "Temperature": 20.6
      }
    }
    yield mqtt_json_sensor(tasmota_sensor_topic, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Define reference data.
    reference = {
      u'time': u'2017-02-16T09:13:52Z',
      u'DS18B20.Temperature': 20.6,
    }

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == reference
    yield record


@pytest_twisted.inlineCallbacks
@pytest.mark.tasmota
@pytest.mark.wemos
def test_tasmota_wemos_dht22(machinery, create_influxdb, reset_influxdb):
    """
    Publish a reading from a Wemos multi sensor device.
    Proof that the reading is processed and stored correctly.
    """

    # Submit a single measurement.
    data = {
      "Time": "2017-10-05T22:39:55",
      "DHT22": {
        "Temperature": 25.4,
        "Humidity": 45
      },
      "TempUnit": "C"
    }
    yield mqtt_json_sensor(tasmota_sensor_topic, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Define reference data.
    reference = {
        u'time': u'2017-10-05T20:39:55Z',
        u'DHT22.Temperature': 25.4,
        u'DHT22.Humidity': 45,
    }

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == reference
    yield record


@pytest_twisted.inlineCallbacks
@pytest.mark.tasmota
@pytest.mark.wemos
def test_tasmota_wemos_multi(machinery, create_influxdb, reset_influxdb):
    """
    Publish a reading from a Wemos multi sensor device.
    Proof that the reading is processed and stored correctly.
    """

    # Submit a single measurement.
    data = {
      "Time": "2017-10-05T22:39:45",
      "DS18x20": {
        "DS1": {
          "Type": "DS18B20",
          "Address": "28FF4CBFA41604C4",
          "Temperature": 25.37
        },
        "DS2": {
          "Type": "DS18B20",
          "Address": "28FF1E7FA116035D",
          "Temperature": 30.44
        },
        "DS3": {
          "Type": "DS18B20",
          "Address": "28FF1597A41604CE",
          "Temperature": 25.81
        }
      },
      "DHT22": {
        "Temperature": 33.2,
        "Humidity": 30
      },
      "TempUnit": "C"
    }
    yield mqtt_json_sensor(tasmota_sensor_topic, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Define reference data.
    reference = {
        u'time': u'2017-10-05T20:39:45Z',
        u'DHT22.Temperature': 33.2,
        u'DHT22.Humidity': 30,
        u'DS18x20.DS1.Temperature': 25.37,
        u'DS18x20.DS2.Temperature': 30.44,
        u'DS18x20.DS3.Temperature': 25.81,
    }

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    logger.info('record: %s', json.dumps(record))
    assert record == reference
    yield record


@pytest_twisted.inlineCallbacks
@pytest.mark.tasmota
def test_tasmota_state(machinery, create_influxdb, reset_influxdb):
    """
    Publish a single STATE reading in Tasmota/JSON format
    to MQTT broker, including a timestamp.
    Proof that the reading is processed and stored correctly.

    https://getkotori.org/docs/handbook/decoders/tasmota.html#submit-example-payload
    """

    # Submit a single measurement.
    data = {
      "Time": "2019-06-02T22:13:07",
      "Uptime": "1T18:10:35",
      "Vcc": 3.182,
      "SleepMode": "Dynamic",
      "Sleep": 50,
      "LoadAvg": 19,
      "Wifi": {
        "AP": 1,
        "SSId": "{redacted}",
        "BSSId": "A0:F3:C1:{redacted}",
        "Channel": 1,
        "RSSI": 100,
        "LinkCount": 1,
        "Downtime": "0T00:00:07"
      }
    }
    yield mqtt_json_sensor(tasmota_state_topic, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Define reference data.
    reference = {
      u'time': u'2019-06-02T20:13:07Z',
      u'Device.Vcc': 3.182,
      u'Device.Sleep': 50,
      u'Device.LoadAvg': 19,
      u'Device.Wifi.Channel': 1,
      u'Device.Wifi.RSSI': 100,
      u'Device.Wifi.LinkCount': 1,
    }

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == reference
    yield record



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import json
import logging

import io
import pandas as pd
import pytest
import pytest_twisted
from twisted.internet import threads

from test.settings.mqttkit import settings, PROCESS_DELAY_MQTT
from test.util import sleep, http_json_sensor, http_get_data
from pandas.testing import assert_frame_equal
from datadiff.tools import assert_equal

logger = logging.getLogger(__name__)


@pytest_twisted.inlineCallbacks
@pytest.mark.http
@pytest.mark.export
def test_export(machinery, create_influxdb, reset_influxdb):
    """
    Submit single reading in JSON format to HTTP API and proof
    it can be retrieved back from the HTTP API in different formats.
    """

    # Submit a single measurement, with timestamp.
    data = {
        'time': 1583810982,
        'temperature': 25.26,
        'humidity': 51.8,
    }
    yield threads.deferToThread(http_json_sensor, settings.channel_path_data, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data is available via HTTP API.
    ts_from = '2020-03-10T00:00:00.000Z'
    ts_to = '2020-03-10T23:59:59.000Z'

    # Reference DataFrame to compare with.
    df_should = pd.DataFrame([[pd.to_datetime("2020-03-10T03:29:42.000000Z"), 51.8, 25.26]], columns=["time", "humidity", "temperature"])

    # CSV format.
    deferred = threads.deferToThread(http_get_data, settings.channel_path_data, format='csv', ts_from=ts_from, ts_to=ts_to)
    yield deferred
    result = pd.read_csv(io.StringIO(deferred.result), parse_dates=["time"])
    assert_frame_equal(result, df_should, check_names=False, check_like=True, check_datetimelike_compat=True)

    # TXT format (same as CSV).
    deferred = threads.deferToThread(http_get_data, settings.channel_path_data, format='txt', ts_from=ts_from, ts_to=ts_to)
    yield deferred
    result = pd.read_csv(io.StringIO(deferred.result), parse_dates=["time"])
    assert_frame_equal(result, df_should, check_names=False, check_like=True, check_datetimelike_compat=True)

    # JSON format.
    deferred = threads.deferToThread(http_get_data, settings.channel_path_data, format='json', ts_from=ts_from, ts_to=ts_to)
    yield deferred
    result = json.loads(deferred.result)
    should = [{"time": "2020-03-10T03:29:42.000Z", "humidity": 51.8, "temperature": 25.26}]
    assert_equal(result, should)

    # XLSX format.
    deferred = threads.deferToThread(http_get_data, settings.channel_path_data, format='xlsx', ts_from=ts_from, ts_to=ts_to)
    yield deferred
    assert deferred.result.startswith(b'PK\x03\x04\x14\x00\x00\x00\x08\x00')

    # HTML format.
    deferred = threads.deferToThread(http_get_data, settings.channel_path_data, format='html', ts_from=ts_from, ts_to=ts_to)
    yield deferred
    assert \
        '<html>' in deferred.result and \
        '<th>temperature</th>' in deferred.result and \
        '<td>25.26</td>' in deferred.result and \
        '<th>humidity</th>' in deferred.result and \
        '<td>51.8</td>' in deferred.result

    # HDF5 format.
    deferred = threads.deferToThread(http_get_data, settings.channel_path_data, format='hdf5', ts_from=ts_from, ts_to=ts_to)
    yield deferred
    assert deferred.result.startswith(b'\x89HDF\r\n\x1a\n\x00\x00\x00\x00\x00')

    import h5py
    from io import BytesIO
    from pandas import array
    hdf = h5py.File(BytesIO(deferred.result), "r")
    assert hdf["/itest_foo_bar/table"].attrs.get("NROWS") == 1
    assert hdf["/itest_foo_bar/table"].attrs.get("index_kind") == b"datetime64"
    assert hdf["/itest_foo_bar/table"].attrs.get("humidity_dtype") == b"float64"
    assert hdf["/itest_foo_bar/table"].attrs.get("temperature_dtype") == b"float64"
    assert hdf["/itest_foo_bar/table"].shape == (1,)
    assert hdf["/itest_foo_bar/table"][()] == array(
        [(1583810982000000000, 51.8, 25.26)],
        dtype=[('index', '<i8'), ('humidity', '<f8'), ('temperature', '<f8')])

    # NetCDF format.
    deferred = threads.deferToThread(http_get_data, settings.channel_path_data, format='nc', ts_from=ts_from, ts_to=ts_to)
    yield deferred
    assert deferred.result.startswith(b'\x89HDF\r\n\x1a\n\x00\x00\x00\x00\x00')

    # Datatables HTML.
    deferred = threads.deferToThread(http_get_data, settings.channel_path_data, format='dt', ts_from=ts_from, ts_to=ts_to)
    yield deferred
    assert b"cdn.datatables.net" in deferred.result



# -*- coding: utf-8 -*-
# (c) 2015-2021 Andreas Motl, <andreas@getkotori.org>
import re

from kotori.daq.strategy import StrategyBase
from kotori.util.common import SmartBunch


class LanStrategy(StrategyBase):

    # Regular expression pattern for decoding MQTT topic address segments.
    pattern = r'^(?P<realm>.+?)/(?P<node>.+?)(?:/(?P<slot>.+?))?$'
    matcher = re.compile(pattern)

    def topic_to_topology(self, topic):
        """
        Decode MQTT topic segments implementing the »basic strategy«.

        The topology hierarchy is directly specified by the MQTT topic and is
        made up of a two path segments::

            realm / node

        The topology identifiers are specified as:

            - "realm" is the designated root realm. You should prefix the topic name
              with this label when opting in for all features of the telemetry platform.
              For other purposes, feel free to publish to any MQTT topic you like.

            - "node" is the node identifier. Choose anything you like. This usually
              gets transmitted from an embedded device node.
        """

        # decode the topic
        m = self.matcher.match(topic)
        if m:
            address = SmartBunch(m.groupdict())
        else:
            address = {}

        return address

    @classmethod
    def topology_to_storage(self, topology, message_type=None):
        """
        Encode topology segment identifiers to database address.

        A database server usually has the concept of multiple databases,
        each with multiple tables. With other databases than RDBMS,
        they might be named differently, but the concept in general
        is the same.

        The topology information (realm, node) will
        get mapped to the database name and measurement name to
        compute the storage location for measurements.

            - realm + node     = database name
            - sensors | events = table name

        """

        # Derive database table suffix from message type.
        table_suffix = self.get_table_suffix(topology, message_type)

        # Use topology information as blueprint for storage address.
        storage = SmartBunch(topology)

        # Format and sanitize all input parameters used for database addressing.
        # Todo: Investigate using tags additionally to / instead of only "storage.measurement".
        sanitize = self.sanitize_db_identifier
        storage.label       = sanitize('{}'.format(storage.node))
        storage.database    = sanitize('{}_{}'.format(storage.realm, storage.node))
        storage.measurement = sanitize('{}'.format(table_suffix))
        storage.measurement_events = sanitize('{}'.format('events'))

        return storage

    @classmethod
    def topology_to_label(cls, topology):
        return '{}-{}'.format(topology.realm, topology.node)



# -*- coding: utf-8 -*-
# (c) 2019-2021 Andreas Motl <andreas@getkotori.org>
from kotori.daq.decoder.airrohr import AirrohrDecoder
from kotori.daq.decoder.tasmota import TasmotaSensorDecoder, TasmotaStateDecoder
from kotori.daq.decoder.schema import MessageType


class DecoderInfo:

    def __init__(self):
        self.message_type = None
        self.decoder = None


class DecoderManager:

    def __init__(self, topology):
        self.topology = topology
        self.info = DecoderInfo()

    def probe(self):

        if 'slot' not in self.topology:
            return False

        # Airrohr
        if self.topology.slot.endswith('airrohr.json'):
            self.info.message_type = MessageType.DATA_CONTAINER
            self.info.decoder = AirrohrDecoder
            return True

        # Tasmota Sensor
        if self.topology.slot.endswith('SENSOR'):
            self.info.message_type = MessageType.DATA_CONTAINER
            self.info.decoder = TasmotaSensorDecoder
            return True

        # Tasmota State
        if self.topology.slot.endswith('STATE'):
            self.info.message_type = MessageType.DATA_CONTAINER
            self.info.decoder = TasmotaStateDecoder
            return True

        return False



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
"""
# conftest.py: sharing fixture functions
https://docs.pytest.org/en/latest/fixture.html#conftest-py-sharing-fixture-functions
"""
import pytest
import logging

import pytest_twisted

from kotori import KotoriBootloader
from test.util import boot_kotori
from test.settings.mqttkit import influx_sensors, influx_events, grafana

logger = logging.getLogger(__name__)


# Disable ellipsis truncation
# https://stackoverflow.com/questions/19171554/preventing-truncation-of-long-strings-in-pytest/60321834#60321834
#from _pytest.assertion import truncate
#truncate.DEFAULT_MAX_LINES = 9999
#truncate.DEFAULT_MAX_CHARS = 9999


def create_machinery(config, scope="package"):

    @pytest_twisted.inlineCallbacks
    @pytest.fixture(scope=scope)
    def machinery():

        # Invoke the machinery.
        logger.info('Spinning up machinery')

        #result = yield threads.deferToThread(boot_kotori)
        result = boot_kotori(config)
        assert isinstance(result, KotoriBootloader)

        #pytest_twisted.blockon(result)
        yield result

    return machinery


machinery = create_machinery('./etc/test/main.ini')
create_influxdb = influx_sensors.make_create_db()
reset_influxdb = influx_sensors.make_reset_measurement()
reset_grafana = grafana.make_reset()

reset_influxdb_events = influx_events.make_reset_measurement()



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

requires = [

    # Core
    'Twisted[tls]==20.3.0',
    'pyOpenSSL>=16.2.0',
    'six>=1.15.0',
    'pyramid==1.9.4',
    'pyramid_jinja2>=2.8,<3',
    'cornice>=5.0.3,<6',
    'simplejson>=3.17.2,<4',
    'Jinja2>=2.11.2,<3',
    'Bunch>=1.0.1,<2',
    'munch>=2.5.0,<3',
    'appdirs>=1.4.3,<2',
    'json-store>=3.0,<4',
    'python-dateutil>=2.8.0,<3',
    'arrow>=0.17.0,<1',
    'funcy>=1.15,<2',
    'attrs>=20.2.0,<21',

    # Bus adapters
    'paho-mqtt>=1.5.1,<2',
    'autobahn[twisted]>=20.7.1,<21',
    'msgpack-python>=0.5.6,<0.6',
    'PyTrie>=0.4.0,<1',

    # Misc
    'distlib>=0.3.1,<0.4',
    'docopt>=0.6.2,<0.7',

    # More dependencies
    'cryptography>=3.1.1',
    'certifi>=2020.6.20',

    'service_identity>=18.1.0,<19',
    'idna>=2.10,<3',
    'pyasn1>=0.4.8,<0.5',
    'pyasn1-modules>=0.2.8,<0.3',

]

extras = {
    'daq': [
        'influxdb>=5.3.0,<6',
        'pytz>=2020.1',
        'requests>=2.12.4,<3',
        'grafana_api_client==0.2.0',
        #'grafana-dashboard-builder==0.1.0a7',      # evaluated, but not suitable
        #'txmongo==16.3.0',
        'pymongo>=3.11.0,<4',
    ],
    'daq_geospatial': [
        'Geohash>=1.0,<2',
        'geopy>=1.12.0,<2',
        'Beaker>=1.9.0,<2',
        'tqdm>=4.19.8,<5',
    ],
    'daq_binary': [
        'pycparser==2.17',          # 2.18
        'pyparsing==2.2.0',
        'pyclibrary==0.1.3',
        'tabulate==0.7.5',          # 0.8.2
        'sympy==0.7.6.1',           # 1.1.1
    ],
    'storage_plus': [
        'alchimia>=0.4,<1',
    ],

    # Data export: Basic formats
    'export': [
        'pyinfluxql>=0.0.1,<1',
        'pandas>=0.25,<1.3',
        'numpy>=1.18,<1.20',
        'XlsxWriter>=1.3.6,<1.4',
    ],

    'plotting': [
        #'dyplot==0.8.8',

        'matplotlib>=3,<3.4',
        #'cairocffi>=0.5.4',
        'bokeh>=1.4.0,<2.3',
        'vincent>=0.4.4,<0.5',
    ],

    # Data export: Scientific data formats like HDF5 and NetCDF and plots from ggplot
    'scientific': [

        # Data
        # ----
        # HDF5
        # "PyTables" requires HDF5 libraries
        # sudo port install hdf5
        'tables>=3.6.1,<3.7',
        'h5py>=2.10.0,<4',

        # NetCDF (Network Common Data Form)
        'xarray>=0.13.0,<0.17',
        # sudo port install netcdf
        'netCDF4>=1.5.4,<1.6',
        #'h5netcdf==0.2.2',

        # Algorithms
        # ----------
        'scipy>=1.4.1,<1.6',
        'ggplot>=0.11.5,<0.12',

        # gfortran
        # aptitude install libatlas-base-dev lapack-dev gfortran or
        # https://gcc.gnu.org/wiki/GFortranBinaries

        # Visualization
        # -------------
        #'seaborn==0.7.1',

    ],

    'firmware': [
        'GitPython>=2.0.5,<3',
        'plumbum>=1.6.1,<1.7',
    ],

}

setup(name='kotori',
      version='0.26.3',
      description='Kotori is a data acquisition, processing and graphing toolkit for humans',
      long_description=README,
      license="AGPL 3, EUPL 1.2",
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Framework :: Pyramid",
        "Framework :: Twisted",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Communications",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Object Brokering",
        "Topic :: System :: Archiving",
        "Topic :: System :: Networking :: Monitoring",
        "Topic :: Text Processing",
        "Topic :: Utilities",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS"
        ],
      author='Andreas Motl',
      author_email='andreas.motl@getkotori.org',
      url='https://github.com/daq-tools/kotori',
      keywords='data acquisition graphing export plotting daq routing engine ' +
               'mqtt http rest amqp wamp sql web html csv json cdf hdf5 png ' +
               'twisted pyramid autobahn influxdb mosquitto grafana mongodb matplotlib ggplot ' +
               'telemetry m2m iot',
      packages=find_packages(),
      include_package_data=True,
      package_data={
        'kotori': [
          'daq/graphing/grafana/resources/*.json',
          'io/export/*.html',
        ],
      },
      zip_safe=False,
      test_suite='kotori.test',
      install_requires=requires,
      extras_require=extras,
      entry_points={
          'console_scripts': [
              'kotori              = kotori:run',
              #'kotori-master      = kotori.master.server:run',
              #'kotori-node        = kotori.node.nodeservice:run',
              #'kotori-wamp-client = kotori.master.client:run_wamp_client',
              'h2m-csv-udp-client  = kotori.vendor.hydro2motion.client:run_udp_client',
              'h2m-csv-udp-fuzzer  = kotori.vendor.hydro2motion.client:run_udp_fuzzer',
              'lst-message         = kotori.vendor.lst.shell:message',
              'kotori-selftest     = kotori.vendor.selftest:run',
          ],
          'paste.app_factory': [
              'main = kotori.frontend.app:main',
          ],
      },
)



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging

import pytest
import pytest_twisted
from twisted.internet import threads

from test.settings.mqttkit import settings, influx_sensors, PROCESS_DELAY_MQTT
from test.util import mqtt_json_sensor, sleep, mqtt_sensor

logger = logging.getLogger(__name__)


@pytest_twisted.inlineCallbacks
@pytest.mark.mqtt
def test_mqtt_to_influxdb_json_single(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single measurement, without timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
    }
    yield threads.deferToThread(mqtt_json_sensor, settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    del record['time']
    assert record == {u'humidity': 83.1, u'temperature': 42.84}
    yield record


@pytest_twisted.inlineCallbacks
@pytest.mark.mqtt
@pytest.mark.legacy
def test_mqtt_to_influxdb_json_legacy_topic(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker on legacy suffix
    and proof it is stored in the InfluxDB database.
    """

    # Submit a single measurement, without timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
    }
    yield threads.deferToThread(mqtt_json_sensor, settings.mqtt_topic_json_legacy, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert 'temperature' in record or 'humidity' in record


@pytest_twisted.inlineCallbacks
@pytest.mark.mqtt
def test_mqtt_to_influxdb_discrete(machinery, create_influxdb, reset_influxdb):
    """
    Publish discrete values to the MQTT broker
    and proof they are stored in the InfluxDB database.
    """

    # Submit discrete measurement values, without timestamp.
    topic_temperature = settings.mqtt_topic_single + '/temperature'
    topic_humidity = settings.mqtt_topic_single + '/humidity'
    yield threads.deferToThread(mqtt_sensor, topic_temperature, 42.84)
    yield threads.deferToThread(mqtt_sensor, topic_humidity, 83.1)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert 'temperature' in record or 'humidity' in record



# -*- coding: utf-8 -*-
# (c) 2020 Andreas Motl <andreas@getkotori.org>

from test.util import InfluxWrapper, GrafanaWrapper

PROCESS_DELAY_MQTT = 0.2


class BasicSettings:

    # InfluxDB settings.
    influx_database = 'basic_node42'
    influx_measurement_sensors = 'sensors'
    influx_measurement_events = 'events'

    # Grafana settings.
    grafana_username = 'admin'
    grafana_password = 'admin'
    grafana_dashboards = ['basic-node42']

    # MQTT channel settings.
    mqtt_topic_single = 'basic/node42/data'
    mqtt_topic_json   = 'basic/node42/data.json'
    mqtt_topic_event  = 'basic/node42/event.json'
    mqtt_topic_homie  = 'basic/node42/data/__json__'
    mqtt_topic_json_legacy = 'basic/node42/message-json'

    # HTTP channel settings.
    channel_path_data    = '/basic/node42/data'
    channel_path_event   = '/basic/node42/event'
    channel_path_airrohr = '/basic/node42/custom/airrohr'


settings = BasicSettings

influx_sensors = InfluxWrapper(database=settings.influx_database, measurement=settings.influx_measurement_sensors)
influx_events = InfluxWrapper(database=settings.influx_database, measurement=settings.influx_measurement_events)
grafana = GrafanaWrapper(settings=settings)

create_influxdb = influx_sensors.make_create_db()
reset_influxdb = influx_sensors.make_reset_measurement()
reset_grafana = grafana.make_reset()



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@getkotori.org>
import logging
import pytest_twisted

from test.settings.mqttkit import settings, influx_sensors, PROCESS_DELAY_MQTT
from test.util import mqtt_json_sensor, sleep

logger = logging.getLogger(__name__)


@pytest_twisted.inlineCallbacks
def test_timestamp_rfc3339(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker,
    using a timestamp in RFC3339 format.
    Proof that the timestamp is processed and stored correctly.
    """

    # Submit a single measurement, with timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
        'timestamp': '2020-03-10 03:38:37.937059000+01:00'
    }
    yield mqtt_json_sensor(settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == {u'time': u'2020-03-10T02:38:37.937059Z', u'humidity': 83.1, u'temperature': 42.84}
    yield record


@pytest_twisted.inlineCallbacks
def test_timestamp_seconds(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker,
    using a timestamp as Unix Epoch in seconds.
    Proof that the timestamp is processed and stored correctly.
    """

    # Submit a single measurement, with timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
        'timestamp': 1583810982
    }
    yield mqtt_json_sensor(settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == {u'time': u'2020-03-10T03:29:42Z', u'humidity': 83.1, u'temperature': 42.84}
    yield record


@pytest_twisted.inlineCallbacks
def test_timestamp_milliseconds(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker,
    using a timestamp as Unix Epoch in milliseconds.
    Proof that the timestamp is processed and stored correctly.
    """

    # Submit a single measurement, with timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
        'timestamp': 1583810982123
    }
    yield mqtt_json_sensor(settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == {u'time': u'2020-03-10T03:29:42.123000Z', u'humidity': 83.1, u'temperature': 42.84}
    yield record


@pytest_twisted.inlineCallbacks
def test_timestamp_microseconds(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker,
    using a timestamp as Unix Epoch in microseconds.
    Proof that the timestamp is processed and stored correctly.
    """

    # Submit a single measurement, with timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
        'timestamp': 1583810982123456
    }
    yield mqtt_json_sensor(settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == {u'time': u'2020-03-10T03:29:42.123456Z', u'humidity': 83.1, u'temperature': 42.84}
    yield record


@pytest_twisted.inlineCallbacks
def test_timestamp_nanoseconds(machinery, create_influxdb, reset_influxdb):
    """
    Publish single reading in JSON format to MQTT broker,
    using a timestamp as Unix Epoch in nanoseconds.
    Proof that the timestamp is processed and stored correctly.
    """

    # Submit a single measurement, with timestamp.
    data = {
        'temperature': 42.84,
        'humidity': 83.1,
        'timestamp': 1583810982123456789
    }
    yield mqtt_json_sensor(settings.mqtt_topic_json, data)

    # Wait for some time to process the message.
    yield sleep(PROCESS_DELAY_MQTT)

    # Proof that data arrived in InfluxDB.
    record = influx_sensors.get_first_record()
    assert record == {u'time': u'2020-03-10T03:29:42.123457Z', u'humidity': 83.1, u'temperature': 42.84}
    yield record
